@@ -1,4 +1,5 @@
 import tblib.pickling_support
+# allows pickling of basic information from tracebacks:
 tblib.pickling_support.install()
 
 import abc
@@ -21,6 +22,12 @@ from pacer.std_logger import get_logger
 
 class DelayedException(object):
 
+    """store exception for throwing it later.
+       we use this for transfering an exception on a worker process to the
+       main process.
+       only works because we imported and installed tblib above
+       """
+
     def __init__(self, ee):
         self.ee = ee
         __,  __, self.tb = sys.exc_info()
@@ -29,19 +36,19 @@ class DelayedException(object):
         raise self.ee, None, self.tb
 
 
-def remote_eval(f_s, args_s):
+def eval_pickled(f_s, args_s):
     """ requires function and args serialized with dill, so we can remote execute even lambdas or
     decorated functions or static methods, which 'pure' multiprocssing can not handle"""
 
     try:
         args = dill.loads(args_s)
     except Exception, e:
-        get_logger(remote_eval).error("got exception when unpickling args: {}".format(e))
+        get_logger(eval_pickled).error("got exception when unpickling args: {}".format(e))
         return Exception(e.message)
     try:
         f = dill.loads(f_s)
     except Exception, e:
-        get_logger(remote_eval).error("got exception when unpickling f: {}".format(e))
+        get_logger(eval_pickled).error("got exception when unpickling f: {}".format(e))
         return Exception(e.message)
     try:
         get_logger(f).info("start {}{} in process {}".format(f.__name__, args, os.getpid()))
@@ -106,18 +113,22 @@ class Engine(object):
         concurrent processing."""
         assert all(isinstance(arg, EnumeratedItem) for arg in enumerated_args)
         new_enumeration, args = unpack(enumerated_args)
+        for arg in args:
+            if isinstance(arg, DelayedException):
+                callback(arg)
+                return
+        # dill can pickle more function types as pickle module (which underlies
+        # multiprocessing) , eg lambdas or static methods, so we transport our computation job
+        # serialized as strings, which we deserialize in the remote process again:
+        f_s = dill.dumps(f)
+        args_s = dill.dumps(args)
+        # we use eval_pickled in both cases to make shure we have the same kind of error handling
+        # for local and remote execution. this supports debugging.
         if Engine.pool is None:
-            f_s = dill.dumps(f)
-            args_s = dill.dumps(args)
-            result = remote_eval(f_s, args_s)
+            result = eval_pickled(f_s, args_s)
             wrap_callback(new_enumeration, callback)(result)
         else:
-            # dill can pickle more function types as pickle module (which underlies
-            # multiprocessing) , eg lambdas or static methods, so we transport our computation job
-            # serialized as strings, which we deserialize in the remote process again:
-            f_s = dill.dumps(f)
-            args_s = dill.dumps(args)
-            Engine.pool.apply_async(remote_eval, (f_s, args_s),
+            Engine.pool.apply_async(eval_pickled, (f_s, args_s),
                                     callback=wrap_callback(new_enumeration, callback))
 
 
@@ -139,14 +150,13 @@ class DataStream(object):
 
     def put(self, item):
         """this is where computation results are passed to."""
-        if isinstance(item.value, DelayedException):
-            item.value.re_raise()
-            item.value.print_tb()
-            raise Exception()
         self._stream.put(item)
 
     def get_all_in_order(self):
         results = [self.get() for _ in range(self.size())]
+        for result in results:
+            if isinstance(result, DelayedException):
+                result.re_raise()
         results.sort(key=lambda en_item: en_item.number)
         return [item.value for item in results]
 
@@ -200,6 +210,7 @@ class Computation(DataStream):
             if isinstance(stream, Computation):
                 stream.start_computations()
         self._start_computation()
+        return self
 
     @abc.abstractmethod
     def compute_size(self):
